/*-----------------------------------------------------------------------
  
  File  : mysort.c
  
Author: Stephan Schulz (schulz@eprover.org)

Contents

  Program that reads and sorts long integers.
  
  Copyright 2015 by the author.

  This code is released under the GNU General Public Licence, version
  2, or, at your choice, any later version. See the file COPYING.
  
Changes

<1> Mon Jun 22 20:43:15 CEST 2015
    New
    
-----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <assert.h>

#define MAXLINE  256
#define MAXITEMS 25000000

long numbers[MAXITEMS];
long tmpnum[MAXITEMS];

/*-----------------------------------------------------------------------
//
// Function: GetUSecTime()
//
//   Return the time in microseconds since the epoch.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

long long GetUSecTime()
{
   struct timeval tv;
   
   gettimeofday(&tv, NULL);
   
   return (long long)tv.tv_sec*1000000ll+tv.tv_usec;
}


/*-----------------------------------------------------------------------
//
// Function: cmpfun()
//
//   Comparison function for qsort()
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

int cmpfun(const void* e1, const void* e2)
{
   const long *p1=e1, *p2=e2;
   if(*p1 > *p2)
   {
      return 1;
   }
   if(*p1 < *p2)
   {
      return -1;
   }
   return 0;
}

/*-----------------------------------------------------------------------
//
// Function: selectionsort() 
//
//   Simple selection sort. Sorts the array in place.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void selectionsort(long arr[], long start, long size)
{
   long tmp, i, j, min_i;

   for(i=start; i<size; i++)
   {
      min_i = i;
      for(j=i+1; j< size; j++)
      {
         if(arr[j]<arr[min_i])
         {
            min_i = j;
         }
      }
      if(min_i != i)
      {
         tmp = arr[i];
         arr[i] = arr[min_i];
         arr[min_i] = tmp;
      }
   }
}


/*-----------------------------------------------------------------------
//
// Function: quicksort() 
//
//   Quicksort implementation.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void quicksort(long arr[], long start, long size)
{
   //printf("qs: %ld %ld\n", start, size);
   if(size > 1)
   {
      long i, j, pi, pivot, tmp, end;
      
      end = start+size-1;
      pi = start+size/2;
      pivot = arr[pi];
      tmp = arr[end];
      arr[end] = arr[pi];
      arr[pi] = tmp;

      j = start; 
      for(i=start; i< end; i++)
      {
         if(arr[i] < pivot)
         {
            tmp = arr[i];
            arr[i] = arr[j];
            arr[j] = tmp;
            j++;
         }
      }
      tmp = arr[end];
      arr[end] = arr[j];
      arr[j] = tmp;
      quicksort(arr, start, j-start);
      quicksort(arr, j+1, end-j);
   }
}


/*-----------------------------------------------------------------------
//
// Function: median()
//
//   Find and returnn the index of the median value of arr[i1],
//   arr[i2], arr[i3].
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

long median(long arr[], long i1, long i2, long i3)
{
   if(arr[i1] < arr[i2])
   {
      if(arr[i2] < arr[i3])
      {
         return i2;
      }
      if(arr[i1]<arr[i3])
      {
         return i3;
      }
      return i1;
   }
   
   if(arr[i1] < arr[i3])
   {
      return i1;
   };
   if(arr[i2]<arr[i3])
   {
         return i3;
   }
   return i2;
}

/*-----------------------------------------------------------------------
//
// Function: quicksortopt() 
//
//   Quicksort implementation with optimisations.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void quicksortopt(long arr[], long start, long size)
{
   //printf("qs: %ld %ld\n", start, size);
   if(size > 3)
   {
      long i, j, pi, pivot, tmp, end;
      
      end = start+size-1;
      pi = median(arr, start, start+size/2, end);
      pivot = arr[pi];
      tmp = arr[end];
      arr[end] = arr[pi];
      arr[pi] = tmp;

      j = start; 
      for(i=start; i< end; i++)
      {
         if(arr[i] < pivot)
         {
            tmp = arr[i];
            arr[i] = arr[j];
            arr[j] = tmp;
            j++;
         }
      }
      tmp = arr[end];
      arr[end] = arr[j];
      arr[j] = tmp;
      quicksortopt(arr, start, j-start);
      quicksortopt(arr, j+1, end-j);
   }
   else if(size > 1)
   {
      selectionsort(arr, start, size);
   }
}

/*-----------------------------------------------------------------------
//
// Function: mrg_sort()
//
//   Mergesort on longs.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void mrg_sort(long arr[], long start, long size)
{
   static long mrgarr[MAXITEMS];

   //printf("mrg: %ld - %ld\n", start, start+size-1);

   if(size > 1)
   {
      long i, a1, a2;

      mrg_sort(arr, start, size/2);
      mrg_sort(arr, start+size/2, size-size/2);

      a1 = start;
      a2 = start+size/2;
      for(i=start; i<start+size; i++)
      {
         if(a1>=start+size/2)
         {
            mrgarr[i] = arr[a2];
            a2++;            
         }
         else if(a2 >= start+size)
         {
            mrgarr[i] = arr[a1];
            a1++;            
         }
         else if(arr[a1]<=arr[a2])
         {
            mrgarr[i] = arr[a1];
            a1++;                        
         }
         else
         {
            mrgarr[i] = arr[a2];
            a2++;                        
         }
      }
      for(i=start; i<start+size; i++)
      {
         arr[i] = mrgarr[i];
      }
   }
}

/*-----------------------------------------------------------------------
//
// Function: bubble_up()
//
//   Perform the bubble-up operation on the given node in the heap
//   represented by arr[0:size].
//
// Global Variables: -
//
// Side Effects    : Changes the array
//
/----------------------------------------------------------------------*/

void bubble_up(long arr[], long size, long node)
{
   long tmp;

   while(node && arr[node]>arr[(node-1)/2])
   {
      tmp = arr[node];
      arr[node] = arr[(node-1)/2];
      arr[(node-1)/2] = tmp;
      node = (node-1)/2;
   }
}
 
/*-----------------------------------------------------------------------
//
// Function: bubble_down()
//
//   Perform the bubble-down operation on the given node in the heap
//   represented by arr[0:size].
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void bubble_down(long arr[], long size, long node)
{
   long tmp, gci, lci, rci;

   while((lci = node*2+1)<size)
   {
      gci = lci;
      rci = lci+1;
      if((rci<size) && (arr[rci]>arr[lci]))
      {
         gci = rci;
      }
      if(arr[gci] <= arr[node])
      {
         break;
      }
      tmp = arr[node];
      arr[node] = arr[gci];
      arr[gci] = tmp;
      node = gci;
   }
}

/*-----------------------------------------------------------------------
//
// Function: heapify()
//
//   Convert arr[0:size] into a binary heap.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void heapify(long arr[], long size)
{
   long last_inner = (size-1)/2, i;

   for(i = last_inner; i>=0; i--)
   {
      bubble_down(arr, size, i);
   }
}

/*-----------------------------------------------------------------------
//
// Function: heap_sort()
//
//   Sort the array using the Heapsort algorithm.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void heap_sort(long arr[], long size)
{
   long i, tmp;

   heapify(arr, size);

   for(i=size-1; i; i--)
   {
      tmp=arr[0];
      arr[0] = arr[i];
      arr[i] = tmp;
      bubble_down(arr, i, 0);
   }
}



/*-----------------------------------------------------------------------
//
// Function: main()
//
//   
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

int main(int argc, char *argv[])
{
   FILE *in = stdin;
   char line[MAXLINE];
   char* inp; 
   long count = 0;
   long long time1, time2;
   
   if(argc > 2)
   {
      fprintf(stderr, "Usage: %s [<file>]\n", argv[0]);
      exit(EXIT_FAILURE);
   }
   if(argc == 2)
   {
      in = fopen(argv[1], "r");
      if(!in)
      {
         perror(argv[0]);
         exit(EXIT_FAILURE);
      }
   }
   while((inp = fgets(line, MAXLINE, in)))
   {
      if(count == MAXITEMS)
      {
         fprintf(stderr,  "Too many lines\n");
         exit(EXIT_FAILURE);
      }
      numbers[count++] = atol(inp);
   }
   if(in!=stdin)
   {
      fclose(in);
   }
   memcpy(tmpnum, numbers, sizeof(numbers));  
   time1 = GetUSecTime();      
   qsort(tmpnum, count, sizeof(long), cmpfun);
   time2 = GetUSecTime();
   printf("qsort time: %lld usec\n", time2-time1);

   memcpy(tmpnum, numbers, sizeof(numbers)); 
   time1 = GetUSecTime(); 
   selectionsort(tmpnum, 0, count);
   time2 = GetUSecTime();
   printf("selsort time: %lld usec\n", time2-time1); 
   
   memcpy(tmpnum, numbers, sizeof(numbers));  
   time1 = GetUSecTime();      
   quicksort(tmpnum, 0, count);
   time2 = GetUSecTime();
   printf("quicksort time: %lld usec\n", time2-time1);

   memcpy(tmpnum, numbers, sizeof(numbers));  
   time1 = GetUSecTime();      
   quicksortopt(tmpnum, 0, count);
   time2 = GetUSecTime();
   printf("quicksortopt time: %lld usec\n", time2-time1);

   memcpy(tmpnum, numbers, sizeof(numbers));  
   time1 = GetUSecTime();      
   mrg_sort(tmpnum, 0, count);
   time2 = GetUSecTime();
   printf("mrgsort time: %lld usec\n", time2-time1);

   memcpy(tmpnum, numbers, sizeof(numbers));  
   time1 = GetUSecTime();      
   heap_sort(tmpnum, count);
   time2 = GetUSecTime();
   printf("heap_sort time: %lld usec\n", time2-time1);   
   
   //for(int i=0; i<count; i++)
   //{
   // printf("%ld\n", tmpnum[i]);
   //}

   exit(EXIT_SUCCESS);
}

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/

